/**
	A generator that infinitely loops around an array
	
	@param array The array to loop around
	@yields Yields items from the array wrapping around to the start.
	@next Decideds if the iterator should go backwards
 */

export function* cyclicArray<T>(array: T[]): Generator<T, never, boolean> {
	let index = 0;
	while (true) {
		const reverse = yield array[index];
		if (reverse) index--;
		else index++;

		if (index < 0) index = array.length - 1;
		else if (index >= array.length) index = 0;
	}
}

/**
	Rotates an array
	
	@param arr The array to rotate
	@param reverse Should the array be rotated in the opposite direction
	@returns The rotated array. For example by passing in [1,2,3] you will recieve [3,1,2]
*/
export const rotateArray = <T>(arr: T[], reverse = false): T[] => {
	if (reverse) arr.unshift(arr.pop()!);
	else arr.push(arr.shift()!);
	return arr;
};

/**
	Ergonomic wrapper for creating new promises
	@returns A tuple of the created promise and it's resolve and reject functions
*/
type resolveCallback<T> = (value: T) => void;
type rejectCallback = (reason?: unknown) => void;
export const newPromise = <T>(): [
	Promise<T>,
	resolveCallback<T>,
	rejectCallback
] => {
	let resolve: resolveCallback<T>;
	let reject: rejectCallback;
	const promise = new Promise(
		(res: resolveCallback<T>, rej: rejectCallback) => {
			resolve = res;
			reject = rej;
		}
	);
	return [promise, resolve!, reject!];
};

//Clone of the https://developer.mozilla.org/en-US/docs/Mozilla/Add-ons/WebExtensions/API/alarms
export namespace Alarms {
	const alarmsMap = new Map<any, [Promise<void>, rejectCallback]>();

	export const create = (delay: number, name?: any) => {
		const alrm = alarmsMap.get(name);
		if (alrm) alrm[1]();

		const [promise, resolve, reject] = newPromise<void>();
		const timeout = setTimeout(resolve, delay);
		promise.catch(() => window.clearTimeout(timeout));

		if (name != undefined) alarmsMap.set(name, [promise, reject]);
		return promise;
	};

	export const get = (name: any) => (alarmsMap.get(name) ?? [])[0];
	export const getAll = () => Array.from(alarmsMap.entries()).map((a) => a[0]);

	export const clear = (name: any) => {
		const alrm = alarmsMap.get(name);
		if (alrm) alrm[1]();
	};
	export const clearAll = () => alarmsMap.forEach((a) => a[1]());
}

/**
	@returns Promise that resolves when an animation on the element finishes 
*/
export const animationPromise = (element: HTMLElement, cancel = true) => {
	const [promise, reject, resolve] = newPromise<void>();
	element.addEventListener("animationend", () => resolve(), true);
	if (cancel) element.addEventListener("animationcancel", () => reject(), true);
	return promise;
};

/**
	Escapes all the keys in the template string
*/
export const html = (strings: TemplateStringsArray, ...keys: string[]) =>
	strings.reduce((p, c, i) => p + escape(keys[i - 1]) + c);

export const escape = (string: string) =>
	string.replace(/</g, "&lt").replace(/>/g, "&gt");
